# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

# -*- coding: utf-8 -*-
"""Module containing classes for VCC - Supplier debug interface.

TODO: Check if all IO parameters in SPMEMSInterfaceRequirements.xlsx defined as safety variables
 automatically have debug switches.
"""

import os
import re

from powertrain_build import build_defs
from powertrain_build.types import byte_size_string, get_bitmask, a2l_range
from powertrain_build.a2l import A2l
from powertrain_build.problem_logger import ProblemLogger


class ExtDbg(ProblemLogger):
    """Class for generating c-files.

    These declares all debug parameters in the VCC - Supplier interface.
    """

    __data_type_size = {'Float32': '4', 'UInt32': '4', 'Int32': '4',
                        'UInt16': '2', 'Int16': '2',
                        'UInt8': '1', 'Int8': '1', 'Bool': '1'}

    def __init__(self, variable_dict, prj_cfg, unit_cfg, integrity_level=build_defs.CVC_ASIL_QM):
        """Constructor.

        Args:
            variable_dict (dict): dictionary with signal information

        Variable dict shall have the following format and is generated by the
        :doc:`CsvSignalInterfaces <signal_interfaces>` class::

            {
                'CAN-Input': {
                    'signal1': {
                        'IOType': 'd',
                        'description': 'Some description',
                        'init': 0,
                        'max': 1,
                        'min': 0,
                        'type': 'UInt8',
                        'unit': '-'
                    },
                    'signal2': {
                        ...
                    }
                },
                'CAN-Output': {
                    'signal3': {
                        ...
                    }
                },
                'xxx-Input': ...,
                'xxx-Output': ...
            }

        The optional keyword arguments may override which code section
        directives to use.
        """
        super().__init__()
        self.set_integrity_level(integrity_level)
        self._prj_cfg = prj_cfg
        self._unit_cfg = unit_cfg
        self._use_volatile_globals = prj_cfg.get_use_volatile_globals()
        self.dbg_dict = self._restruct_data(variable_dict)

    @staticmethod
    def _restruct_data(variable_dict):
        """Restructure input variables per data-type.

        This will be used for declaring the variables and generating the A2L-file.
        """
        data = {'outputs': {}, 'inputs': {}}
        for inp in variable_dict.keys():
            if re.match(r'.*Output$', inp) is not None:
                iotype = 'outputs'
            else:
                iotype = 'inputs'
            for var, var_data in variable_dict[inp].items():
                data_type_size = byte_size_string(var_data['type'])
                data[iotype].setdefault(data_type_size, {})[var] = var_data
        return data

    def set_integrity_level(self, integrity_level):
        """Set integrity level of code generation.

        Args:
            integrity_level (str): integrity level of the unit from 'A' to 'D' or 'QM'
        """
        self._disp_start = integrity_level['DISP']['START']
        self._disp_end = integrity_level['DISP']['END']
        self._cal_start = integrity_level['CAL']['START']
        self._cal_end = integrity_level['CAL']['END']
        self._code_start = integrity_level['CODE']['START']
        self._code_end = integrity_level['CODE']['END']

    def _a2l_dict(self, var_dict, function):
        """Generate dict defining parameters for a2l-generation."""
        def _range(data):
            range_a2l = a2l_range(data['type'])
            if data['min'] == '-':
                a2l_min = range_a2l[0]
            else:
                a2l_min = data['min']
            if data['max'] == '-':
                a2l_max = range_a2l[1]
            else:
                a2l_max = data['max']
            return a2l_min, a2l_max

        res = {'vars': {},
               'function': function}
        for var, data in self._type_order_iterator(var_dict):
            resv = res['vars']
            a2l_min, a2l_max = _range(data)
            a2l_data = {
                'bitmask': get_bitmask(data['type']),
                'description': data['description'],
                'lsb': '2^0',
                'max': a2l_max,
                'min': a2l_min,
                'offset': '0',
                'unit': '',
                'x_axis': None,
                'y_axis': None
            }
            var_db = f'c{var[1:]}_db'
            var_sw = self._var_name_to_dbgsw_name(var)
            resv.setdefault(var_db, {})['a2l_data'] = a2l_data
            a2l_data = {
                'bitmask': get_bitmask(data['type']),
                'description': f'debug switch for {var_db} (1=bdsw act)',
                'lsb': '2^0',
                'max': '1',
                'min': '0',
                'offset': '0',
                'unit': '',
                'x_axis': None,
                'y_axis': None
            }
            resv.setdefault(var_sw, {})['a2l_data'] = a2l_data
            resv[var_db]['array'] = []
            resv[var_sw]['array'] = []
            resv[var_db]['function'] = [function]
            resv[var_sw]['function'] = [function]
            resv[var_db]['var'] = {'cvc_type': 'CVC_CAL',
                                   'type': data['type'],
                                   'var': var}
            resv[var_sw]['var'] = {'cvc_type': 'CVC_CAL',
                                   'type': 'Bool',
                                   'var': var}
        return res

    @staticmethod
    def _var_name_to_dbgsw_name(name):
        """Convert a variable name to a debug switch name."""
        # the below conversion would generate a correct name for the
        # debug switch, however the current build systemt generates one
        # like the currently returned name
        # return re.sub(r'\w(\w+?)_\w+?_(\w+)', r'c\1_B_\2_sw', name)
        return re.sub(r'\w(\w+)', r'c\1_sw', name)

    @staticmethod
    def _type_order_iterator(var_items):
        """Get iterator over all variables.

        In data type size order, and then in alphabetical order.
        """
        for _, typ_data in var_items.items():
            for var in sorted(typ_data.keys()):
                yield (var, typ_data[var])

    def _gen_dbg_c_file(self, data, filename):
        """Generate debug c-files.

        These define all the debug labels for
        the supplier input and output signals.
        """
        with open(filename, 'w', encoding="utf-8") as fh_c:
            fh_c.write(self._unit_cfg.base_types_headers)
            fh_c.write('#define CVC_DISP\n')

            # define extrern variable references
            fh_c.write(f'#include "{self._disp_start}"\n')
            for var, var_data in self._type_order_iterator(data):
                fh_c.write(f"extern CVC_DISP {var_data['type']} {var};\n")
            fh_c.write(f'#include "{self._disp_end}"\n\n')

            # define debug calibration constants
            fh_c.write(f'#include "{self._cal_start}"\n')
            fh_c.write('\n/* Debug values */\n\n')
            for var, var_data in self._type_order_iterator(data):
                initial_value = var_data['min'] if var_data['min'] != "-" and float(var_data['min']) > 0 else "0"
                if self._use_volatile_globals:
                    fh_c.write(f"volatile {var_data['type']} c{var[1:]}_db = {initial_value};\n")
                else:
                    fh_c.write(f"{var_data['type']} c{var[1:]}_db = {initial_value};\n")

            fh_c.write('\n/* Debug switches */\n\n')
            for var, var_data in self._type_order_iterator(data):
                sw_name = self._var_name_to_dbgsw_name(var)
                if self._use_volatile_globals:
                    fh_c.write(f"volatile Bool {sw_name} = 0;\n")
                else:
                    fh_c.write(f"Bool {sw_name} = 0;\n")
            fh_c.write(f'#include "{self._cal_end}"\n\n')

            # set the variable to the debug calibration constants
            fh_c.write('/***********************/\n')
            fh_c.write('/* debug functionality */\n')
            fh_c.write('/***********************/\n\n')
            _, fname_tmp = os.path.split(filename)
            func_name = fname_tmp.split('.')[-2]
            fh_c.write(f'#include "{self._code_start}"\n')
            fh_c.write(f'void {func_name}(void) {{\n')
            for var, var_data in self._type_order_iterator(data):
                sw_name = self._var_name_to_dbgsw_name(var)
                fh_c.write(f'  if ({sw_name}) {{\n')
                fh_c.write(f'    {var} = c{var[1:]}_db;\n  }}\n')
            fh_c.write(f'}}\n#include "{self._code_end}"\n\n')
            self.info('Generated %s', filename)

    def gen_dbg_files(self, file_path_inputs, file_path_outputs):
        """Generate the c-files and A2L-files.

        These declares all the supplier interface debug parameters and functions.

        Args:
            file_path_inputs (str): path to the debug inputs c-file.
            file_path_outputs (str): path to the debug outputs c-file.
        """
        _, file_name_inputs = os.path.split(file_path_inputs)
        _, file_name_outputs = os.path.split(file_path_outputs)

        if not self.dbg_dict['inputs']:
            self.info(f"Skipping {file_name_inputs} as there were no corresponding vars.")
        else:
            self._gen_dbg_c_file(self.dbg_dict['inputs'], file_path_inputs + '.c')
            a2l_dict_in = self._a2l_dict(self.dbg_dict['inputs'], file_name_inputs)
            a2l = A2l(a2l_dict_in, self._prj_cfg)
            a2l.gen_a2l(file_path_inputs + '.a2l')

        if not self.dbg_dict['outputs']:
            self.info(f"Skipping {file_name_outputs} as there were no corresponding vars.")
        else:
            self._gen_dbg_c_file(self.dbg_dict['outputs'], file_path_outputs + '.c')
            a2l_dict_out = self._a2l_dict(self.dbg_dict['outputs'], file_name_outputs)
            a2l = A2l(a2l_dict_out, self._prj_cfg)
            a2l.gen_a2l(file_path_outputs + '.a2l')



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Module containing classes for the signal interfaces report."""
from collections import defaultdict
from powertrain_build.html_report import HtmlReport


class SigIfHtmlReport(HtmlReport):
    """Signal interface html report.

    Inherits class :doc:`HtmlReport <html_report>`.
    """

    __intro = """<h2 class="nocount">Table of contents</h2>
    <ol>
      <li><a href="#introduction">Introduction</a></li>
      <li><a href="#signal_details">Detailed signal information</a></li>
      <li><a href="#signal_index">Signal index</a></li>
    </ol>
    <h2 id="introduction">Introduction</h2>
    <p>This documents lists the signal interfaces.</p>
    """

    __table_unused = """
  <table id="unused">
    <thead>
      <tr>
        <th>Variable</th>
        <th>Configurations</th>
      </tr>
    </thead>
    <tbody>
"""

    __table_detailed_sig_desc = """
  <table id="sig_desc">
    <thead>
      <tr>
        <th>Unit</th>
        <th>Type</th>
        <th>Class</th>
        <th>Min</th>
        <th>Max</th>
        <th>Lsb</th>
        <th>Offset</th>
        <th>Configs</th>
      </tr>
    </thead>
    <tbody>
"""

    __table_detailed_sig_desc_ext = """
  <table id="sig_desc">
    <thead>
      <tr>
        <th>Unit</th>
        <th>Type</th>
        <th>Init Value</th>
        <th>Min</th>
        <th>Max</th>
        <th>Config Doc</th>
      </tr>
    </thead>
    <tbody>
"""

    __table_detailed_sig_def = """
  <table id="detailed_sig_def">
    <thead>
      <tr>
        <th>Defined in unit</th>
        <th>in projects</th>
      </tr>
    </thead>
    <tbody>
"""

    __table_detailed_sig_cons = """
  <table id="detailed_sig_cons">
    <thead>
      <tr>
        <th>Consumed by units</th>
        <th>in projects</th>
      </tr>
    </thead>
    <tbody>
"""

    def __init__(self, build_cfg, unit_cfg, sig_if):
        """Initialize class instance.

        Args:
            build_cfg (BuildProjConfig): holds all the build configs
            unit_cfg (UnitConfigs): object holding all unit interfaces
            sig_if (SignalInterfaces): object holding signal interface information
        """
        super().__init__('Signal interface report')
        self._build_cfg = build_cfg
        self._unit_cfg = unit_cfg
        self._sig_if = sig_if
        self._format_unit_data()

    def _gen_sigs_details(self):
        """Generate detailed signal information."""
        out = '  <h2 id="signal_details">Detailed Signal Information</h2>\n'
        for signal in sorted(self._if.keys()):
            out += self._gen_sig_details(signal)
        return out

    def _gen_sig_details(self, signal):
        """Generate detailed signal information."""
        out = f'  <h3 id="{signal}">{signal}</h3>\n'
        unit = tuple(self._if[signal]['def'].keys())[0]  # ugly! use the first defining unit.
        # Check that it is a unit, and not from the inteface definition
        if unit[:2] == 'Vc':
            sig_def = defaultdict(lambda: '-', self._unit_cfg.get_per_unit_cfg()[unit]['outports'][signal])
            out += f'  <p>{sig_def["description"]}</p>'
            out += self.__table_detailed_sig_desc
            out += '      <tr>\n'
            out += f'        <td>{sig_def["unit"]}</td>\n'
            out += f'        <td>{sig_def["type"]}</td>\n'
            out += f'        <td>{sig_def["class"]}</td>\n'
            out += f'        <td>{sig_def["min"]}</td>\n'
            out += f'        <td>{sig_def["max"]}</td>\n'
            out += f'        <td>{sig_def["lsb"]}</td>\n'
            out += f'        <td>{sig_def["offset"]}</td>\n'
            out += f'        <td>{sig_def["configs"]}</td>\n'
            out += '      </tr>\n'
            out += '    </tbody>\n'
            out += '  </table>\n  <p></p>\n'
        else:
            sig_def = defaultdict(lambda: '-', self._sig_if.get_raw_io_cnfg()[unit][signal])
            out += f'  <p>{sig_def["description"]}</p>'
            out += self.__table_detailed_sig_desc_ext
            out += '      <tr>\n'
            out += f'        <td>{sig_def["unit"]}</td>\n'
            out += f'        <td>{sig_def["type"]}</td>\n'
            out += f'        <td>{sig_def["init"]}</td>\n'
            out += f'        <td>{sig_def["min"]}</td>\n'
            out += f'        <td>{sig_def["max"]}</td>\n'
            out += f'        <td>{unit}</td>\n'
            out += '      </tr>\n'
            out += '    </tbody>\n'
            out += '  </table>\n  <p></p>\n'

        out += self.__table_detailed_sig_def
        for unit, prjs in self._if[signal]['def'].items():
            out += '      <tr>\n'
            out += f'        <td>{unit}</td>\n'
            out += f'        <td>{prjs}</td>\n'
            out += '      </tr>\n'
        out += '    </tbody>\n'
        out += '  </table>\n'

        if 'consumers' in self._if[signal]:
            out += '  <p></p>\n'
            out += self.__table_detailed_sig_cons
            for units, prjs in self._if[signal]['consumers'].items():
                out += '      <tr>\n'
                out += f'        <td>{units}</td>\n'
                out += f'        <td>{prjs}</td>\n'
                out += '      </tr>\n'
            out += '    </tbody>\n'
            out += '  </table>\n'
        return out

    def _format_unit_data(self):
        """Format data per unit."""
        self._if = {}
        prj_name = self._build_cfg.get_prj_config()
        # internal signals
        prj_cfg = self._unit_cfg.get_per_cfg_unit_cfg()
        ext_out_prj = self._sig_if.get_external_outputs()
        if 'outports' in prj_cfg:
            for sig, data in prj_cfg['outports'].items():
                prod_unit = list(data.keys())[0]
                self._if.setdefault(sig, {}).setdefault('def', {}).setdefault(prod_unit, []).append(prj_name)
                cons_units = []
                for sig_type, osig in ext_out_prj.items():
                    if sig in osig:
                        cons_units = [sig_type]
                        break
                if sig in prj_cfg['inports']:
                    cons_units.extend(list(prj_cfg['inports'][sig].keys()))
                if cons_units:
                    (self._if[sig].setdefault('consumers', {})
                                  .setdefault(tuple(sorted(cons_units)), [])
                                  .append(prj_name))

        # external signals
        prj_cfg = self._unit_cfg.get_per_cfg_unit_cfg()
        ext_in_prj = self._sig_if.get_external_inputs()
        for sig_type, osig in ext_in_prj.items():
            for sig, data in osig.items():
                self._if.setdefault(sig, {}).setdefault('def', {}).setdefault(sig_type, []).append(prj_name)
                if 'inports' in prj_cfg and sig in prj_cfg['inports']:
                    (self._if[sig].setdefault('consumers', {})
                                  .setdefault(tuple(sorted(prj_cfg['inports'][sig].keys())), [])
                                  .append(prj_name))

    def _gen_signal_toc(self):
        """Generate a signal TOC for the unit with signal inconsistencies.

        Hyperlinks to more in depth signal information.
        """
        out = '  <h2 id="signal_index">Signal index</h2>\n'
        for signal in sorted(self._if.keys()):
            out += f'  <div><a href="#{signal}">{signal}</a></div>\n'
        return out

    def gen_contents(self):
        """Generate report contents from the signal interfaces data.

        Overrides HtmlReport.gen_contents()
        """
        html = []
        html += self.__intro
        html += self._gen_sigs_details()
        html += self._gen_signal_toc()
        return ''.join(html)



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Create a python package for powertrain-build using setuptools and PBR."""

from setuptools import setup

setup(
    setup_requires=['pbr'],
    pbr=True,
)



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Python module used for abstracting Hardware Abstraction Layer specifications"""
import re
from pathlib import Path
from ruamel.yaml import YAML
from powertrain_build.interface.csp_api import CspApi
from powertrain_build.lib import logger

LOGGER = logger.create_logger('base')


def get_hal_list(app):
    """Get translated hal name from yaml file

    Args:
        app (Application): Pybuild project
    Returns:
        cmake (str): a string contains translated hal name
    """
    translation_files = app.get_translation_files()
    hala = HALA(app)
    hala.parse_definition(translation_files)
    hal_translations = hala.get_map()

    hals = set()
    for definitions in hala.translations.values():
        for definition in definitions:
            hal_abbreviation = definition[hala.position.api.value]
            real_hal_name = get_real_hal_name(hal_abbreviation, hal_translations)
            hals.add((hal_abbreviation, real_hal_name))

    cmake = ""
    for hal_abbreviation, real_hal_name in hals:
        lib = re.sub('-', '_', f'hal_{hal_abbreviation}' + '_libhal_' + real_hal_name).upper()
        include = re.sub('-', '_', f'hal_{hal_abbreviation}' + '_include_dir').upper()
        cmake += f"LIST(APPEND extra_libraries ${{{lib}}})\n"
        cmake += f"LIST(APPEND EXTRA_INCLUDE_DIRS ${{{include}}})\n"
    return cmake


def strip_hal_name(hal_name):
    """Strip hal name

    Args:
        hal_name (str): hal name
    Returns:
        (str): stripped hal name
    """
    return hal_name.replace('_hal', '')


def verify_name(hal_name, api_map):
    """Verify hal name

    Args:
        hal_name (str): hal name
        api_map (dict): hal translation map
    """
    if strip_hal_name(hal_name) not in api_map:
        raise HalTranslationException(
            f"{hal_name} does not exist in the hal translation file."
        )


def get_real_hal_name(hal_name, translation_map):
    """Get real hal name from translation map file.

    Args:
        hal_name (str): hal abreviation

    Returns:
        real_hal_name (str): real name of a hal
    """
    verify_name(hal_name, translation_map)
    return translation_map.get(strip_hal_name(hal_name))


class UnknownAccessorError(Exception):
    """Error when setting a producer and there already exists one"""
    def __init__(self, hal, signal, accessor):
        """Set error message

        Args:
            hal (HAL): Hal where the problem is
            signal (Signal): Signal with the problem
            accessor (str): Unknown accessor type
        """
        super().__init__()
        self.message = f"Accessor of type {accessor} for {signal.name} in {hal.name} is not handled"


class HalTranslationException(Exception):
    """Class for hal translation exceptions"""


class HALA(CspApi):
    """Hardware abstraction layer abstraction"""

    def __repr__(self):
        """String representation of HALA"""
        return (f"<HALA {self.name}"
                f" app_side insignals: {len(self.signal_names['app']['insignals'])}"
                f" app_side outsignals: {len(self.signal_names['app']['outsignals'])}>")

    def get_map(self):
        """Get hal translation map

        Returns:
            (dict): hal translation map
        """
        path = self.get_map_file()
        if path.is_file():
            return self._get_hal_translation(path)
        return {}

    @staticmethod
    def get_map_file():
        """Get hal translation map file

        Returns:
            (Path): hal translation map file
        """
        return Path("Projects", "CSP", "hal_list.yaml")

    def get_api_name(self, api_name):
        real_hal_name = get_real_hal_name(
            api_name,
            self.api_map
        )
        return real_hal_name

    def verify_api(self, api_name):
        verify_name(api_name, self.api_map)

    @staticmethod
    def _get_hal_translation(path):
        """Get translated hal names

        Args:
            path (Path): path to the hal translation list.
        Returns:
            hal_translation_content (dict): translated hal names
        """
        hal_translation_content = None
        if path and path.is_file():
            with path.open("r") as file_handler:
                yaml = YAML(typ="safe", pure=True)
                hal_translation_content = yaml.load(file_handler)

        if not hal_translation_content:
            if hal_translation_content is None:
                raise HalTranslationException(
                    "No hal translation file given."
                )
            if isinstance(hal_translation_content, dict):
                raise HalTranslationException(
                    "Hal translation file are empty."
                )
            raise HalTranslationException("Bad hal translation format.")
        return hal_translation_content

    def check_endpoints(self):
        pass

    def extract_endpoint_definitions(self, raw):
        """Extract endpoint definitions from raw data

        Args:
            (dict): endpoint definitions
        """
        self.parse_api_definitions(raw.get("hal", {}))

    @staticmethod
    def extract_definition(definition):
        """Extract definition from hal

        Args:
            definition (dict): hal definition

        Returns:
            (dict): hal definition
        """
        if isinstance(definition, list):
            specifications = {
                'hals': definition
            }
        else:
            specifications = {
                'properties': definition.get('properties', []),
                'methods': definition.get('methods', [])
            }
        return specifications



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Module containing classes for generation of signal consistency check report."""
from powertrain_build.html_report import HtmlReport


class SigConsHtmlReportBase(HtmlReport):
    """Generate html report from the signal consistency check result.

    (see :doc:`signal_interfaces`)

    Inherits :doc:`HtmlReport <html_report>`.
    """

    __intro = """<h2 class="nocount">Table of contents</h2>
    <ol>
      <li><a href="#introduction">Introduction</a></li>
      <li><a href="#unit_details">Detailed unit information</a></li>
      <li><a href="#ext_sigs">External signals</a>
      <ol>
        <li><a href="#ext_missing">Missing external signals</a></li>
        <li><a href="#ext_unused">Unused external signals</a></li>
      </ol></li>
      <li><a href="#unit_index">Unit index</a></li>
    </ol>
    <h2 id="introduction">Introduction</h2>
    <p>This documents lists inconsistencies in the internal and external signal configuration.</p>
    """
    _table_unused = """
  <table id="unused">
    <thead>
      <tr>
        <th>Variable</th>
        <th>Configurations</th>
      </tr>
    </thead>
    <tbody>
    """
    _table_incons = """
  <table id="inconsistent">
    <thead>
      <tr>
        <th>Variable</th>
        <th>Variable parameter</th>
        <th>Difference</th>
      </tr>
    </thead>
    <tbody>
    """
    _table_unknown = """
          <table id="unknown">
            <thead>
              <tr>
                <th>Variable</th>
                <th>Units</th>
              </tr>
            </thead>
            <tbody>
        """

    def __init__(self, res_dict=None):
        """Initialize class instance.

        Args:
            res_dict (dict): result dict from the signal interface consistency check

        The dict shall have the following format::

            {
                "sigs": { "ext": {"missing": {},
                                  "unused": {},
                                  "inconsistent_defs": {}},
                          "int": {"UNIT_NAME": {"missing": {},
                                                "unused": {},
                                                "multiple_defs": {}
                                                "inconsistent_defs": {}}
                }
            }
        """
        super().__init__('Signal consistency report')

    @staticmethod
    def _set_to_str(set_):
        """Convert a set to a sorted string."""
        out = ""
        for conf in sorted(set_):
            out += conf + ', '
        return out[:-2]

    def _gen_missing_sigs(self, unit_data, out=''):
        """Generate the unit specific information for missing signal in a unit."""
        out += self._table_unused
        for var in sorted(unit_data['missing'].keys()):
            configs = unit_data['missing'][var]
            out += '  <tr>\n'
            out += f'        <td>{var}</td>\n'
            out += f'        <td>{self._set_to_str(configs)}</td>\n'
            out += '      </tr>\n'
        out += '    </tbody>\n'
        out += '  </table>\n'
        return out

    def _gen_unused_sigs(self, unit_data, out=''):
        """Generate the unit specific information for the unused signals wihtin a unit."""
        out += self._table_unused
        for var in sorted(unit_data['unused'].keys()):
            configs = unit_data['unused'][var]
            out += '  <tr>\n'
            out += f'        <td>{var}</td>\n'
            out += f'        <td>{self._set_to_str(configs)}</td>\n'
            out += '      </tr>\n'
        out += '    </tbody>\n'
        out += '  </table>\n'
        return out

    def _gen_multiple_def_sigs(self, unit_data, out=''):
        """Generate unit specific information for the signals that are generated more than once."""
        out += self._table_unused
        for var in sorted(unit_data['multiple_defs'].keys()):
            configs = unit_data['multiple_defs'][var]
            out += '  <tr>\n'
            out += f'        <td>{var}</td>\n'
            out += f'        <td>{self._set_to_str(configs)}</td>\n'
            out += '      </tr>\n'
        out += '    </tbody>\n'
        out += '  </table>\n'
        return out

    def _gen_ext_inconsistent_defs(self, unit, out=''):
        """Generate a report of inconsistent variable definition parameters.

        Report inconsistencies between the producing signal definition, and the
        signal definitions in the external interface definition.
        """
        out += self._table_incons
        incons_unit = self._res_dict['sigs']['ext']['inconsistent_defs'][unit]
        for var in sorted(incons_unit.keys()):
            first_cells = f'  <tr>\n        <td>{var}</td>\n'
            for v_par, desc in incons_unit[var].items():
                out += first_cells
                out += f'        <td>{v_par}</td>\n'
                out += f'        <td>{desc}</td>\n'
                out += '      </tr>\n'
                first_cells = '      <tr>\n      <td></td>\n'
        out += '    </tbody>\n'
        out += '  </table>\n'
        return out

    def _gen_int_inconsistent_defs(self, unit_data, out=''):
        """Generate a report of inconsistent variable definition parameters.

        Inconsistent for between the producing signal definition, and the
        consuming signal definitions for SPM internal signals.
        """
        out += self._table_incons
        for var in sorted(unit_data['inconsistent_defs'].keys()):
            configs = unit_data['inconsistent_defs'][var]
            first_cells = f'  <tr>\n        <td>{var}</td>\n'
            for v_par, desc in configs.items():
                out += first_cells
                out += f'        <td>{v_par}</td>\n'
                out += f'        <td>{desc}</td>\n'
                out += '      </tr>\n'
                first_cells = '      <tr>\n      <td></td>\n'
        out += '    </tbody>\n'
        out += '  </table>\n'
        return out



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Module containing classes for generation of signal consistency check report."""
import json

from powertrain_build.signal_incons_html_rep_base import SigConsHtmlReportBase


class SigConsHtmlReportAll(SigConsHtmlReportBase):
    """Generate html report from the signal consistency check result.

    (see :doc:`signal_interfaces`)

    Inherits :doc:`HtmlReport <html_report>`.
    """

    __intro_all = """<h2 id="introduction">Introduction</h2>
    <p>This documents lists inconsistencies in the internal and external signal configuration.</p>
"""
    __table_incons = """  <table id="inconsistent">
    <thead>
      <tr>
        <th>Variable</th>
        <th>Variable parameter</th>
        <th>Difference</th>
        <th>Configurations</th>
      </tr>
    </thead>
    <tbody>"""

    __toc_ext_sig = """<h2 class="nocount">Table of contents</h2>'
    <ol>
    <li><a href="#ext_sigs">External signals</a>
    <ol>
    <li><a href="#ext_missing">Missing external signals</a></li>
    <li><a href="#ext_unused">Unused external signals</a></li>
    </ol></li>
    <h2 id="ext_sigs">Signals missing and unused in the interface definition</h2>\n
    """

    _toc_unit_details = """<h2 class="nocount">Table of contents</h2>
    <ol>
    <li><a href="#unit_details">Detailed unit information</a></li>
    <li><a href="#unit_index">Unit index</a></li>
    </ol>
    """

    def __init__(self, res_dicts=None):
        """Initialize class instance.

        Args:
            res_dict (dict): result dict from the signal interface consistency check

        The dict shall have the following format::

            {
                "sigs": { "ext": {"missing": {},
                                  "unused": {},
                                  "inconsistent_defs": {}},
                          "int": {"UNIT_NAME": {"missing": {},
                                                "unused": {},
                                                "multiple_defs": {}
                                                "inconsistent_defs": {}}
                },
                "never_active_signals": {
                    "UNIT_NAME": [signal_one, ...]
                }
            }
        """
        super().__init__(res_dicts)
        self.set_res_dict(res_dicts)

    def _gen_unit_toc(self):
        """Generate a unit TOC for the unit with signal inconsistencies.

        Hyperlinks to more in depth unit information.
        """
        self._out_all_unit_toc = {}
        for prj, units in self._all_units.items():
            out = '  <h2 id="unit_index">Unit index</h2>\n'
            for unit in sorted(units):
                out += f'  <div><a href="#{unit}">{unit}</a></div>\n'
            self._out_all_unit_toc.update({prj: out})

    def _gen_units(self):
        """Generate all the information regarding all the units."""
        self._out_all_units = {}
        self._prj = ''
        for prj, units in self._all_units.items():
            out = '  <h2 id="unit_details">Detailed Unit Information</h2>\n'
            for unit in sorted(units):
                out += self._gen_unit(prj, unit)
            self._out_all_units.update({prj: out})
        return self._out_all_units

    def _gen_unit(self, project, unit):
        """Generate the html-report for the unit specific information."""
        unit_data_all = {}
        out = f'  <h3 id="{unit}">{unit}</h3>\n'
        if project in self._int_units_all and unit in self._int_units_all[project]:
            for prj, res_dict in self._res_dict_all.items():
                if unit in res_dict['sigs']['int']:
                    unit_data = res_dict['sigs']['int'][unit]
                    unit_data_all.update({prj: unit_data})
            _res_dict = self._res_dict_all.get(project)
            unit_data = _res_dict['sigs']['int'][unit]
            out += '  <h4>Missing signals</h4>\n' \
                   '  <p>Inports whose signals are not generated in the ' \
                   'listed configuration(s).</p>'
            out += self._gen_missing_sigs(unit_data, unit_data_all)
            out += '  <h4>Unused signals</h4>\n' \
                   '  <p>Outports that are generated, but not used in the ' \
                   'listed configuration(s).</p>'
            out += self._gen_unused_sigs(unit_data, unit_data_all)
            out += '  <h4>Multiple defined signals</h4>\n' \
                   '  <p>Outports that are generated more than once in ' \
                   'the listed configuration(s).</p>'
            out += self._gen_multiple_def_sigs(unit_data, unit_data_all)
            out += '  <h4>Internal signal inconsistencies</h4>\n' \
                   '  <p>Inports that have different variable definitions ' \
                   'than the producing outport.</p>'
            out += self._gen_int_inconsistent_defs(unit_data, unit_data_all)
        if project in self._ext_units_all and unit in self._ext_units_all[project]:
            out += self._gen_unit_ext(project, unit)
        if project in self._res_dict_all and \
            'never_active_signals' in self._res_dict_all[project] and \
                unit in self._res_dict_all[project]['never_active_signals']:
            out += '  <h4>Never active signals</h4>\n' \
                   '  <p>Never active signals will not appear in generated .c file, ' \
                   'signals probablty lead to terminators in Simulink model.</p>'
            out += self._gen_never_active_sigs(self._res_dict_all[project]['never_active_signals'][unit])
        return out

    def _gen_unit_ext(self, project, unit):
        out = ''
        self._unit_data_all = {}
        if unit in self._res_dict_all[project]['sigs']['ext']['inconsistent_defs'].keys():
            out += '  <h4>External signal inconsistencies</h4>\n' \
                   '  <p>In-/Out-ports that have different variable definitions ' \
                   'than in the interface definition file.</p>'
            out += self._gen_ext_inconsistent_defs(project, unit)
        return out

    def _gen_signals_table(self, unit_data, unit_data_all, key, out=''):
        """Generate the unit specific information for KEY in a unit."""

        if key not in unit_data:
            return out

        out += self._table_unused
        for var in sorted(unit_data[key]):
            configs_str = ""
            for unit_data_cfg in unit_data_all.values():
                if key in unit_data_cfg and var in unit_data_cfg[key]:
                    configs = unit_data_cfg[key][var]
                    configs_str += f"  {self._set_to_str(configs)}"
            out += '  <tr>\n'
            out += f'        <td>{var}</td>\n'
            out += f'        <td>{configs_str}</td>\n'
            out += '      </tr>\n'
        out += '    </tbody>\n'
        out += '  </table>\n'
        return out

    def _gen_missing_sigs(self, unit_data, unit_data_all, out=''):
        """Generate the unit specific information for missing signal in a unit."""
        return self._gen_signals_table(unit_data, unit_data_all, key='missing', out=out)

    def _gen_unused_sigs(self, unit_data, unit_data_all, out=''):
        """Generate the unit specific information for the unused signals wihtin a unit."""
        return self._gen_signals_table(unit_data, unit_data_all, key='unused', out=out)

    def _gen_multiple_def_sigs(self, unit_data, unit_data_all, out=''):
        """Generate unit specific information for the signals that are generated more than once."""
        return self._gen_signals_table(unit_data, unit_data_all, key='multiple_defs', out=out)

    def _gen_never_active_sigs(self, never_active_signals):
        """Generate unit specific information for the signals that are never active."""
        out = self._table_unused
        for signal in never_active_signals:
            out += '  <tr>\n'
            out += f'        <td>{signal}</td>\n'
            out += '        <td></td>\n'
            out += '      </tr>\n'
        out += '    </tbody>\n'
        out += '  </table>\n'
        return out

    def _gen_ext_inconsistent_defs(self, project, unit, out=''):
        """Generate a report of inconsistent variable definition parameters.

        Report inconsistencies between the producing signal definition, and the
        signal definitions in the external interface definition.
        """
        inconsistent_defs_key = 'inconsistent_defs'
        if inconsistent_defs_key not in self._res_dict_all[project]['sigs']['ext']:
            return out

        out += self.__table_incons
        incons_unit = self._res_dict_all[project]['sigs']['ext'][inconsistent_defs_key][unit]
        for var in sorted(incons_unit.keys()):
            first_cells = f'\n      <tr>\n        <td>{var}</td>\n'
            for v_par, desc in incons_unit[var].items():
                out += first_cells
                out += f'        <td>{v_par}</td>\n'
                out += f'        <td>{desc}</td>\n'
                out += f'        <td>{project}</td>\n'
                out += '      </tr>\n'
                first_cells = '      <tr>\n      <td></td>\n'
        out += '    </tbody>\n'
        out += '  </table>\n'
        return out

    def _gen_int_inconsistent_defs(self, unit_data, unit_data_all, out=''):
        """Generate a report of inconsistent variable definition parameters.

        Inconsistent for between the producing signal definition, and the
        consuming signal definitions for SPM internal signals.
        """
        inconsistent_defs_key = 'inconsistent_defs'
        if inconsistent_defs_key not in unit_data:
            return out

        out += self.__table_incons
        for var in sorted(unit_data[inconsistent_defs_key].keys()):
            configs_str = ""
            for prj_cfg in unit_data_all.keys():
                configs_str += f"  {prj_cfg}"
            first_cells = f'\n      <tr>\n        <td>{var}</td>\n'
            configs = unit_data[inconsistent_defs_key][var]
            for v_par, desc in configs.items():
                out += first_cells
                out += f'        <td>{v_par}</td>\n'
                out += f'        <td>{desc}</td>\n'
                out += f'        <td>{configs_str}</td>\n'
                out += '      </tr>\n'
                first_cells = '      <tr>\n      <td></td>\n'
        out += '    </tbody>\n'
        out += '  </table>\n'
        return out

    def _gen_ext_signals_report(self, type_, comment):
        """Generate report for external signals."""
        out = f'  <h3 id="ext_{type_}">{type_.capitalize()} external signals</h3>\n'
        out += f'<p>{comment}</p>'
        try:
            res_dict = self._res_dict_all.get(self._prj)
            out += self._table_unused
            ext_data = res_dict['sigs']['ext'][type_]
            for var in sorted(ext_data.keys()):
                configs_str = ""
                for res_dict_cfg in self._res_dict_all.values():
                    if type_ in res_dict_cfg['sigs']['ext']:
                        ext_data_cfg = res_dict_cfg['sigs']['ext'][type_]
                        configs = ext_data_cfg[var]
                        configs_str += "  " + self._set_to_str(configs)
                out += '  <tr>\n'
                out += f'        <td>{var}</td>\n'
                out += f'        <td>{configs_str}</td>\n'
                out += '      </tr>\n'
            out += '    </tbody>\n'
            out += '  </table>\n'
        except KeyError:
            pass
        return out

    def set_res_dict(self, res_dicts):
        """Set the result dictionary used to generate the html-report.

        Args:
            res_dicts (dict): result dict from the signal interface consistency check

        See class constructor for dict structure.
        """
        # nesting defaultdicts is bad so we use this hack
        regular_res_dicts = json.loads(json.dumps(res_dicts))
        self._res_dict_all = regular_res_dicts
        self._ext_units_all = {}
        self._int_units_all = {}
        self._all_units = {}
        for prj, res_dict in self._res_dict_all.items():
            _ext_units = set()
            _int_units = set()
            _units_with_never_active_signals = set()
            if res_dict is not None and 'sigs' in res_dict:
                if 'ext' in res_dict['sigs'] and 'inconsistent_defs' in res_dict['sigs']['ext']:
                    _ext_units = set(res_dict['sigs']['ext']['inconsistent_defs'].keys())
                    self._ext_units_all.update({prj: _ext_units})
                if 'int' in res_dict['sigs']:
                    _int_units = set(res_dict['sigs']['int'].keys())
                    self._int_units_all.update({prj: _int_units})
                if 'never_active_signals' in res_dict:
                    _units_with_never_active_signals = res_dict['never_active_signals'].keys()
            self._all_units[prj] = _ext_units | _int_units | _units_with_never_active_signals

    def _gen_contents_intro(self):
        """Generate report contents from the signal interfaces data."""
        html = []
        html += self.__intro_all
        return ''.join(html)

    def _gen_contents_toc_ext_sig(self):
        """Generate report contents from the signal interfaces data.

        Specialises HtmlReport.gen_contents()
        """
        html = []
        html += self.__toc_ext_sig
        return ''.join(html)

    def _gen_contents_toc_unit_details(self):
        """Generate report contents from the signal interfaces data.

        Specialises HtmlReport.gen_contents()
        """
        html = []
        html += self._toc_unit_details
        return ''.join(html)

    def _gen_contents_ext_signals(self):
        """Generate report contents from the signal inconsistency check result dictionary."""
        html = []
        html += self._gen_contents_toc_ext_sig()
        html += self._gen_ext_signals_report('missing', 'Signals not generated by '
                                                        'Vcc SW, but are defined in '
                                                        'the Interface definition to be '
                                                        'generated')
        html += self._gen_ext_signals_report('unused', 'Signals defined to be generated by '
                                                       'supplier SW, but are not used '
                                                       'by VCC SW')
        return ''.join(html)

    def _generate_report_string_intro(self):
        """Generate a html report as string."""
        html = []
        html += self._gen_header()
        html += self._gen_contents_intro()
        html += self._gen_end()
        return ''.join(html)

    def _generate_report_string_ext_signals(self):
        """Generate a html report as string."""
        html = []
        html += self._gen_header()
        html += self._gen_contents_ext_signals()
        html += self._gen_end()
        return ''.join(html)

    def generate_report_file_signal_check(self, filename):
        """Generate a html report and save to file."""
        filename_intro = filename + '_intro.html'
        with open(filename_intro, 'w', encoding="utf-8") as fhndl:
            fhndl.write(self._generate_report_string_intro())

        self._gen_units()
        self._gen_unit_toc()
        for key, out in self._out_all_units.items():
            html = []
            html += self._gen_header()
            html += self._gen_contents_toc_unit_details()
            html += out
            html += self._out_all_unit_toc.get(key, '')
            html += self._gen_end()
            with open(f'{filename}_{key}.html', 'w', encoding="utf-8") as fhndl:
                fhndl.write(''.join(html))



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Module for extraction Energy Management System"""
import os

from powertrain_build.interface.base import BaseApplication, Signal
from powertrain_build.lib import logger
from powertrain_build.signal_interfaces import CsvSignalInterfaces
from powertrain_build.build_proj_config import BuildProjConfig

LOGGER = logger.create_logger(__file__)


class CsvEMS(BaseApplication, CsvSignalInterfaces):
    """Supplier part of the ECM"""
    def __init__(self):
        self._signals = {}
        self.interfaces = {}
        self._insignals = None
        self._outsignals = None
        self.projects = {}

    def parse_definition(self, definition):
        """Read the interface files.

        Args:
            definition (Path): Path to ProjectCfg.json
        """
        self.build_cfg = BuildProjConfig(os.path.normpath(str(definition)))
        CsvSignalInterfaces.__init__(self, self.build_cfg, [])
        self.projects[self.build_cfg.name] = self.build_cfg
        self.config_path = self.build_cfg.get_if_cfg_dir()
        self.name = self.build_cfg.name  # Set name for CsvSignalInterfaces
        self._parse_io_cnfg(self.config_path)
        self.name = 'Supplier_' + self.build_cfg.name  # set name for BaseApplication

    def _get_signals(self):
        """Look through interfaces and create Signal objects"""
        self._insignals = set()
        self._outsignals = set()
        for interface_name, data in self.interfaces.items():
            for signal_name, signal_data in data.items():
                if signal_data['IOType'] == '-':
                    # Signal is inactive
                    continue
                if signal_name in self._signals:
                    signal = self._signals[signal_name]
                else:
                    signal = Signal(signal_name, self)
                    self._signals.update({signal_name: signal})
                if 'output' in interface_name.lower():
                    # outport from ECM. Inport in EMS (from SPM)
                    self._insignals.add(signal_name)
                    signal.consumers = interface_name
                else:
                    signal.producers = interface_name
                    self._outsignals.add(signal_name)

    def get_signal_properties(self, signal):
        """Find properties for a signal from interface files"""
        for interface_name, interface_data in self.interfaces.items():
            if signal.name in interface_data.keys():
                properties = interface_data[signal.name]
                properties['interface'] = interface_name
                return properties
        return {}



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Module to export information of global variables from powertrain_build projects."""

import argparse
import os
import sys
from typing import Dict, List, Optional, Tuple

from ruamel.yaml import YAML

from powertrain_build.build_proj_config import BuildProjConfig
from powertrain_build.feature_configs import FeatureConfigs
from powertrain_build.unit_configs import UnitConfigs


PARSER_HELP = "Export global variables."


def get_global_variables(project_config_path: str) -> Dict:
    """Get global variables connected to PyBuild project.

    Args:
        project_config_path (str): Path to ProjectCfg.json file.
    Returns:
        (Dict): Dict containing project name and its global variables (name, type).
    """
    project_name, project_unit_config = _get_project_data(project_config_path)

    variable_types = ["outports", "local_vars", "calib_consts", "nvm"]
    variables = []
    for variable_type in variable_types:
        if variable_type not in project_unit_config:
            continue

        variables_info = [
            {"name": variable, "type": _get_variable_type(variable_info)}
            for variable, variable_info in project_unit_config[variable_type].items()
        ]
        variables.extend(variables_info)
    return {"name": project_name, "variables": variables}


def _get_variable_type(variable_info: Dict) -> str:
    """Get variable type from variable info.

    Args:
        variable_info (Dict): Dictionary with the variable info.

    Returns:
        str: Variable type.
    """
    unit_name = list(variable_info.keys())[0]  # Getting any unit name, since variable type should be the same
    return variable_info[unit_name]["type"]


def _get_project_data(project_config_path: str) -> Tuple[str, Dict]:
    """Gets data for a powertrain-build project.

    Args:
        project_config_path (str): Path to ProjectCfg.json file.
    Returns:
        project_name (str): Name of PyBuild project.
        project_unit_config (dict): Dict mapping variable types to variables and their data.
    """
    build_cfg = BuildProjConfig(os.path.normpath(project_config_path))
    feature_cfg = FeatureConfigs(build_cfg)
    unit_cfg = UnitConfigs(build_cfg, feature_cfg)
    project_unit_config = unit_cfg.get_per_cfg_unit_cfg()
    return build_cfg.name, project_unit_config


def _export_yaml(data: Dict, file_path: str) -> None:
    """Exports data from dictionary to a yaml file.

    Args:
        data (Dict): Dictionary with data.
        file_path (str): Path of the file to export data.
    """
    with open(file_path, "w", encoding="utf-8") as yaml_file:
        yaml = YAML()
        yaml.default_flow_style = False
        yaml.dump(data, yaml_file)


def export_global_vars(args: argparse.Namespace):
    """Exports global variables as yaml file."""
    global_variables = get_global_variables(args.project_config)
    _export_yaml(global_variables, args.output_file)


def _main(argv: Optional[List[str]] = None):
    """Main function for CLI."""
    parser = argparse.ArgumentParser(description=PARSER_HELP)
    configure_parser(parser)
    args = parser.parse_args(argv)
    args.func(args)


def configure_parser(parser: argparse.ArgumentParser):
    """Configures the argument parser for the script.

    Args:
        parser (argparse.ArgumentParser): Argument parser.
    """
    parser.add_argument(
        "--project-config",
        help="Project root configuration file.",
        required=True,
    )
    parser.add_argument(
        "--output-file",
        help="Output file to export global variables.",
        required=True,
    )
    parser.set_defaults(func=export_global_vars)


if __name__ == "__main__":
    sys.exit(_main(sys.argv[1:]))



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Module for various helper functions."""
import json
import collections
from pathlib import Path
from subprocess import getoutput


def get_repo_root():
    """ Return absolute path to repository where script is executed, regardless
        of the current script's location.

        Returns:
            path (str): Absolute, canonical path to the top-level repository.
                        if not a git repository, returns current working dir

    """
    try:
        root = Path(getoutput('git rev-parse --show-toplevel')).resolve()
    except (FileNotFoundError, OSError):
        root = Path.cwd().resolve()
    return str(root)


def create_dir(path: Path):
    """If the directory for a given directory path does not exist, create it.
    Including parent directories.

    Args:
        path (Path): Path to directory.
    Returns:
        path (Path): Path to directory.
    """
    if not path.is_dir():
        path.mkdir(parents=True)
    return path


def merge_dicts(dict1, dict2, handle_collision=lambda a, b: a, merge_recursively=False):
    """Merge two dicts.

    Args:
        dict1 (dict): dict to merge
        dict2 (dict): dict to merge
        handle_collision (function(arg1, arg2)): function to resolve key collisions,
            default keeps original value in dict1
        merge_recursively (bool): if set to True it merges nested dicts recursively with handle_collision
            resolving collisions with non-dict types.
    Returns:
        dict: the result of the two merged dicts
    """
    result = dict1.copy()
    for key, value in dict2.items():
        if key not in result:
            result.update({key: value})
        elif isinstance(result[key], dict) and \
                isinstance(value, dict) and \
                merge_recursively:
            result[key] = merge_dicts(result[key], value, handle_collision, merge_recursively)
        else:
            result[key] = handle_collision(result[key], value)
    return result


def deep_dict_update(base, add):
    """Recursively update a dict that may contain sub-dicts.

    Args:
        base (dict): The base dict will be updated with the contents
                     of the add dict
        add (dict): This dict will be added to the base dict

    Returns:
        dict: the updated base dict is returned

    """
    for key, value in add.items():
        if key not in base:
            base[key] = value
        elif isinstance(value, dict):
            deep_dict_update(base[key], value)
    return base


def deep_json_update(json_file, dict_to_merge):
    """ Recursively update a json file with the content of a dict.

    Args:
        json_file (path): json file.
        dict_to_merge (dict): Dictionary that will be merged into json file, overwriting values and adding keys.
    """
    with open(json_file, 'r', encoding="utf-8") as fle:
        json_dict = json.load(fle)
    merged_dict = merge_dicts(
        json_dict,
        dict_to_merge,
        handle_collision=lambda a, b: b,
        merge_recursively=True
    )
    with open(json_file, 'w', encoding="utf-8") as fle:
        json.dump(merged_dict, fle, indent=2)


def recursive_default_dict():
    """Returns recursively defined default dict. This allows people to insert
       arbitrarily complex nested data into the dict without getting KeyErrors.

    Returns:
        defaultdict(self): A new defaultdict instance, recursively defined.
    """
    return collections.defaultdict(recursive_default_dict)


def to_normal_dict(weird_dict):
    """Converts nested dict to normal, suitable for YAML/JSON dumping.

    Args:
        weird_dict (dict): Any dict-like item that can be converted to a dict.
    Returns:
        dict(nested): An identical nested dict structure, but using real dicts.
    """
    for key, value in weird_dict.items():
        if isinstance(value, dict):
            weird_dict[key] = to_normal_dict(value)
    return dict(weird_dict)



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Defines header file names for different sections generated c-code."""

CVC_CODE_START = 'CVC_CODE_START.h'  # Header file name for CVC code start
CVC_CODE_END = 'CVC_CODE_END.h'  # Header file name for CVC code end
CVC_CODE_ASIL_A_START = 'CVC_CODE_ASIL_A_START.h'  # Header file name for CVC code start
CVC_CODE_ASIL_A_END = 'CVC_CODE_ASIL_A_END.h'  # Header file name for CVC code end
CVC_CODE_ASIL_B_START = 'CVC_CODE_ASIL_B_START.h'  # Header file name for CVC code start
CVC_CODE_ASIL_B_END = 'CVC_CODE_ASIL_B_END.h'  # Header file name for CVC code end
CVC_CODE_ASIL_C_START = 'CVC_CODE_ASIL_C_START.h'  # Header file name for CVC code start
CVC_CODE_ASIL_C_END = 'CVC_CODE_ASIL_C_END.h'  # Header file name for CVC code end
CVC_CODE_ASIL_D_START = 'CVC_CODE_ASIL_D_START.h'  # Header file name for CVC code start
CVC_CODE_ASIL_D_END = 'CVC_CODE_ASIL_D_END.h'  # Header file name for CVC code end

CVC_DISP_START = 'CVC_DISP_START.h'  # Header file name for CVC measurable variables start
CVC_DISP_END = 'CVC_DISP_END.h'  # Header file name for CVC measurable variables end
CVC_DISP_ASIL_A_START = 'CVC_DISP_ASIL_A_START.h'  # Header file name for CVC ASIL A measurable variables start
CVC_DISP_ASIL_A_END = 'CVC_DISP_ASIL_A_END.h'  # Header file name for CVC ASIL A measurable variables end
CVC_DISP_ASIL_B_START = 'CVC_DISP_ASIL_B_START.h'  # Header file name for CVC ASIL B measurable variables start
CVC_DISP_ASIL_B_END = 'CVC_DISP_ASIL_B_END.h'  # Header file name for CVC ASIL B measurable variables end
CVC_DISP_ASIL_C_START = 'CVC_DISP_ASIL_C_START.h'  # Header file name for CVC ASIL C measurable variables start
CVC_DISP_ASIL_C_END = 'CVC_DISP_ASIL_C_END.h'  # Header file name for CVC ASIL C measurable variables end
CVC_DISP_ASIL_D_START = 'CVC_DISP_ASIL_D_START.h'  # Header file name for CVC ASIL D measurable variables start
CVC_DISP_ASIL_D_END = 'CVC_DISP_ASIL_D_END.h'  # Header file name for CVC ASIL D measurable variables end
CVC_DISP_SEC_START = 'CVC_DISP_SEC_START.h'  # Header file name for CVC secure measurable variables start
CVC_DISP_SEC_END = 'CVC_DISP_SEC_END.h'  # Header file name for CVC secure measurable variables end

CVC_CAL_START = 'CVC_CAL_START.h'  # Header file name for CVC calibration data start
CVC_CAL_END = 'CVC_CAL_END.h'  # Header file name for CVC calibration data end
CVC_CAL_ASIL_A_START = 'CVC_CAL_ASIL_A_START.h'  # Header file name for CVC calibration data start
CVC_CAL_ASIL_A_END = 'CVC_CAL_ASIL_A_END.h'  # Header file name for CVC calibration data end
CVC_CAL_ASIL_B_START = 'CVC_CAL_ASIL_B_START.h'  # Header file name for CVC calibration data start
CVC_CAL_ASIL_B_END = 'CVC_CAL_ASIL_B_END.h'  # Header file name for CVC calibration data end
CVC_CAL_ASIL_C_START = 'CVC_CAL_ASIL_C_START.h'  # Header file name for CVC calibration data start
CVC_CAL_ASIL_C_END = 'CVC_CAL_ASIL_C_END.h'  # Header file name for CVC calibration data end
CVC_CAL_ASIL_D_START = 'CVC_CAL_ASIL_D_START.h'  # Header file name for CVC calibration data start
CVC_CAL_ASIL_D_END = 'CVC_CAL_ASIL_D_END.h'  # Header file name for CVC calibration data end

CVC_CONST_START = 'CVC_CONST_START.h'  # Header file name for CVC calibration data start
CVC_CONST_END = 'CVC_CONST_END.h'  # Header file name for CVC calibration data end
CVC_CONST_ASIL_A_START = 'CVC_CONST_ASIL_A_START.h'  # Header file name for CVC calibration data start
CVC_CONST_ASIL_A_END = 'CVC_CONST_ASIL_A_END.h'  # Header file name for CVC calibration data end
CVC_CONST_ASIL_B_START = 'CVC_CONST_ASIL_B_START.h'  # Header file name for CVC calibration data start
CVC_CONST_ASIL_B_END = 'CVC_CONST_ASIL_B_END.h'  # Header file name for CVC calibration data end
CVC_CONST_ASIL_C_START = 'CVC_CONST_ASIL_C_START.h'  # Header file name for CVC calibration data start
CVC_CONST_ASIL_C_END = 'CVC_CONST_ASIL_C_END.h'  # Header file name for CVC calibration data end
CVC_CONST_ASIL_D_START = 'CVC_CONST_ASIL_D_START.h'  # Header file name for CVC calibration data start
CVC_CONST_ASIL_D_END = 'CVC_CONST_ASIL_D_END.h'  # Header file name for CVC calibration data end

CVC_NVM_START = 'CVC_NVM_START.h'  # Header file name for CVC Non Volatile Memory start
CVC_NVM_END = 'CVC_NVM_END.h'  # Header file name for CVC Non Volatile Memory end
CVC_NVM_P_START = 'CVC_NVM_P_START.h'  # Header file name for persistent Non Volatile Memory start
CVC_NVM_P_END = 'CVC_NVM_P_END.h'  # Header file name for persistent Non Volatile Memory end

PREDECL_CODE_START = 'PREDECL_CODE_START.h'  # Header file name for CVC code start
PREDECL_CODE_END = 'PREDECL_CODE_END.h'  # Header file name for CVC code end
PREDECL_CODE_ASIL_A_START = 'PREDECL_CODE_ASIL_A_START.h'  # Header file name for CVC code start
PREDECL_CODE_ASIL_A_END = 'PREDECL_CODE_ASIL_A_END.h'  # Header file name for CVC code end
PREDECL_CODE_ASIL_B_START = 'PREDECL_CODE_ASIL_B_START.h'  # Header file name for CVC code start
PREDECL_CODE_ASIL_B_END = 'PREDECL_CODE_ASIL_B_END.h'  # Header file name for CVC code end
PREDECL_CODE_ASIL_C_START = 'PREDECL_CODE_ASIL_C_START.h'  # Header file name for CVC code start
PREDECL_CODE_ASIL_C_END = 'PREDECL_CODE_ASIL_C_END.h'  # Header file name for CVC code end
PREDECL_CODE_ASIL_D_START = 'PREDECL_CODE_ASIL_D_START.h'  # Header file name for CVC code start
PREDECL_CODE_ASIL_D_END = 'PREDECL_CODE_ASIL_D_END.h'  # Header file name for CVC code end

PREDECL_DISP_START = 'PREDECL_DISP_START.h'  # Header file name for CVC measurable variables start
PREDECL_DISP_END = 'PREDECL_DISP_END.h'  # Header file name for CVC measurable variables end
PREDECL_DISP_ASIL_A_START = 'PREDECL_DISP_ASIL_A_START.h'  # Header file name for CVC ASIL A measurable variables start
PREDECL_DISP_ASIL_A_END = 'PREDECL_DISP_ASIL_A_END.h'  # Header file name for CVC ASIL A measurable variables end
PREDECL_DISP_ASIL_B_START = 'PREDECL_DISP_ASIL_B_START.h'  # Header file name for CVC ASIL B measurable variables start
PREDECL_DISP_ASIL_B_END = 'PREDECL_DISP_ASIL_B_END.h'  # Header file name for CVC ASIL B measurable variables end
PREDECL_DISP_ASIL_C_START = 'PREDECL_DISP_ASIL_C_START.h'  # Header file name for CVC ASIL C measurable variables start
PREDECL_DISP_ASIL_C_END = 'PREDECL_DISP_ASIL_C_END.h'  # Header file name for CVC ASIL C measurable variables end
PREDECL_DISP_ASIL_D_START = 'PREDECL_DISP_ASIL_D_START.h'  # Header file name for CVC ASIL D measurable variables start
PREDECL_DISP_ASIL_D_END = 'PREDECL_DISP_ASIL_D_END.h'  # Header file name for CVC ASIL D measurable variables end
PREDECL_DISP_SEC_START = 'PREDECL_DISP_SEC_START.h'  # Header file name for CVC secure measurable variables start
PREDECL_DISP_SEC_END = 'PREDECL_DISP_SEC_END.h'  # Header file name for CVC secure measurable variables end

PREDECL_CAL_START = 'PREDECL_CAL_START.h'  # Header file name for CVC calibration data start
PREDECL_CAL_END = 'PREDECL_CAL_END.h'  # Header file name for CVC calibration data end
PREDECL_CAL_ASIL_A_START = 'PREDECL_CAL_ASIL_A_START.h'  # Header file name for CVC calibration data start
PREDECL_CAL_ASIL_A_END = 'PREDECL_CAL_ASIL_A_END.h'  # Header file name for CVC calibration data end
PREDECL_CAL_ASIL_B_START = 'PREDECL_CAL_ASIL_B_START.h'  # Header file name for CVC calibration data start
PREDECL_CAL_ASIL_B_END = 'PREDECL_CAL_ASIL_B_END.h'  # Header file name for CVC calibration data end
PREDECL_CAL_ASIL_C_START = 'PREDECL_CAL_ASIL_C_START.h'  # Header file name for CVC calibration data start
PREDECL_CAL_ASIL_C_END = 'PREDECL_CAL_ASIL_C_END.h'  # Header file name for CVC calibration data end
PREDECL_CAL_ASIL_D_START = 'PREDECL_CAL_ASIL_D_START.h'  # Header file name for CVC calibration data start
PREDECL_CAL_ASIL_D_END = 'PREDECL_CAL_ASIL_D_END.h'  # Header file name for CVC calibration data end
PREDECL_CAL_EXT_START = 'PREDECL_CAL_EXT_START.h'  # Header file name for CVC calibration data start
PREDECL_CAL_EXT_END = 'PREDECL_CAL_EXT_END.h'  # Header file name for CVC calibration data end
PREDECL_CAL_MERG_START = 'PREDECL_CAL_MERG_START.h'  # Header file name for CVC calibration data start
PREDECL_CAL_MERG_END = 'PREDECL_CAL_MERG_END.h'  # Header file name for CVC calibration data end

PREDECL_CONST_START = 'PREDECL_CONST_START.h'  # Header file name for CVC calibration data start
PREDECL_CONST_END = 'PREDECL_CONST_END.h'  # Header file name for CVC calibration data end
PREDECL_CONST_ASIL_A_START = 'PREDECL_CONST_ASIL_A_START.h'  # Header file name for CVC calibration data start
PREDECL_CONST_ASIL_A_END = 'PREDECL_CONST_ASIL_A_END.h'  # Header file name for CVC calibration data end
PREDECL_CONST_ASIL_B_START = 'PREDECL_CONST_ASIL_B_START.h'  # Header file name for CVC calibration data start
PREDECL_CONST_ASIL_B_END = 'PREDECL_CONST_ASIL_B_END.h'  # Header file name for CVC calibration data end
PREDECL_CONST_ASIL_C_START = 'PREDECL_CONST_ASIL_C_START.h'  # Header file name for CVC calibration data start
PREDECL_CONST_ASIL_C_END = 'PREDECL_CONST_ASIL_C_END.h'  # Header file name for CVC calibration data end
PREDECL_CONST_ASIL_D_START = 'PREDECL_CONST_ASIL_D_START.h'  # Header file name for CVC calibration data start
PREDECL_CONST_ASIL_D_END = 'PREDECL_CONST_ASIL_D_END.h'  # Header file name for CVC calibration data end
PREDECL_CONST_EXT_START = 'PREDECL_CONST_EXT_START.h'  # Header file name for CVC calibration data start
PREDECL_CONST_EXT_END = 'PREDECL_CONST_EXT_END.h'  # Header file name for CVC calibration data end
PREDECL_CONST_MERG_START = 'PREDECL_CONST_MERG_START.h'  # Header file name for CVC calibration data start
PREDECL_CONST_MERG_END = 'PREDECL_CONST_MERG_END.h'  # Header file name for CVC calibration data end

PREDECL_NVM_START = 'PREDECL_NVM_START.h'  # Header file name for CVC Non Volatile Memory start
PREDECL_NVM_END = 'PREDECL_NVM_END.h'  # Header file name for CVC Non Volatile Memory end
PREDECL_NVM_P_START = 'PREDECL_NVM_P_START.h'  # Header file name for persistent Non Volatile Memory start
PREDECL_NVM_P_END = 'PREDECL_NVM_P_END.h'  # Header file name for persistent Non Volatile Memory end

PREDECL_START = 'PREDECL_START.h'
PREDECL_END = 'PREDECL_END.h'

CVC_CODE_QM = {'START': CVC_CODE_START, 'END': CVC_CODE_END}
CVC_CODE_A = {'START': CVC_CODE_ASIL_A_START, 'END': CVC_CODE_ASIL_A_END}
CVC_CODE_B = {'START': CVC_CODE_ASIL_B_START, 'END': CVC_CODE_ASIL_B_END}
CVC_CODE_C = {'START': CVC_CODE_ASIL_C_START, 'END': CVC_CODE_ASIL_C_END}
CVC_CODE_D = {'START': CVC_CODE_ASIL_D_START, 'END': CVC_CODE_ASIL_D_END}

CVC_DISP_QM = {'START': CVC_DISP_START, 'END': CVC_DISP_END}
CVC_DISP_A = {'START': CVC_DISP_ASIL_A_START, 'END': CVC_DISP_ASIL_A_END}
CVC_DISP_B = {'START': CVC_DISP_ASIL_B_START, 'END': CVC_DISP_ASIL_B_END}
CVC_DISP_C = {'START': CVC_DISP_ASIL_C_START, 'END': CVC_DISP_ASIL_C_END}
CVC_DISP_D = {'START': CVC_DISP_ASIL_D_START, 'END': CVC_DISP_ASIL_D_END}
CVC_DISP_SECURE = {'START': CVC_DISP_SEC_START, 'END': CVC_DISP_SEC_END}

CVC_CAL_QM = {'START': CVC_CAL_START, 'END': CVC_CAL_END}
CVC_CAL_A = {'START': CVC_CAL_ASIL_A_START, 'END': CVC_CAL_ASIL_A_END}
CVC_CAL_B = {'START': CVC_CAL_ASIL_B_START, 'END': CVC_CAL_ASIL_B_END}
CVC_CAL_C = {'START': CVC_CAL_ASIL_C_START, 'END': CVC_CAL_ASIL_C_END}
CVC_CAL_D = {'START': CVC_CAL_ASIL_D_START, 'END': CVC_CAL_ASIL_D_END}

CVC_CONST_QM = {'START': CVC_CONST_START, 'END': CVC_CONST_END}
CVC_CONST_A = {'START': CVC_CONST_ASIL_A_START, 'END': CVC_CONST_ASIL_A_END}
CVC_CONST_B = {'START': CVC_CONST_ASIL_B_START, 'END': CVC_CONST_ASIL_B_END}
CVC_CONST_C = {'START': CVC_CONST_ASIL_C_START, 'END': CVC_CONST_ASIL_C_END}
CVC_CONST_D = {'START': CVC_CONST_ASIL_D_START, 'END': CVC_CONST_ASIL_D_END}

CVC_NVM = {'START': CVC_NVM_START, 'END': CVC_NVM_END}
CVC_NVM_P = {'START': CVC_NVM_P_START, 'END': CVC_NVM_P_END}

PREDECL = {'START': PREDECL_START, 'END': PREDECL_END}

PREDECL_CODE_QM = {'START': PREDECL_CODE_START, 'END': PREDECL_CODE_END}
PREDECL_CODE_A = {'START': PREDECL_CODE_ASIL_A_START, 'END': PREDECL_CODE_ASIL_A_END}
PREDECL_CODE_B = {'START': PREDECL_CODE_ASIL_B_START, 'END': PREDECL_CODE_ASIL_B_END}
PREDECL_CODE_C = {'START': PREDECL_CODE_ASIL_C_START, 'END': PREDECL_CODE_ASIL_C_END}
PREDECL_CODE_D = {'START': PREDECL_CODE_ASIL_D_START, 'END': PREDECL_CODE_ASIL_D_END}

PREDECL_DISP_QM = {'START': PREDECL_DISP_START, 'END': PREDECL_DISP_END}
PREDECL_DISP_A = {'START': PREDECL_DISP_ASIL_A_START, 'END': PREDECL_DISP_ASIL_A_END}
PREDECL_DISP_B = {'START': PREDECL_DISP_ASIL_B_START, 'END': PREDECL_DISP_ASIL_B_END}
PREDECL_DISP_C = {'START': PREDECL_DISP_ASIL_C_START, 'END': PREDECL_DISP_ASIL_C_END}
PREDECL_DISP_D = {'START': PREDECL_DISP_ASIL_D_START, 'END': PREDECL_DISP_ASIL_D_END}
PREDECL_DISP_SECURE = {'START': PREDECL_DISP_SEC_START, 'END': PREDECL_DISP_SEC_END}

PREDECL_CAL_QM = {'START': PREDECL_CAL_START, 'END': PREDECL_CAL_END}
PREDECL_CAL_A = {'START': PREDECL_CAL_ASIL_A_START, 'END': PREDECL_CAL_ASIL_A_END}
PREDECL_CAL_B = {'START': PREDECL_CAL_ASIL_B_START, 'END': PREDECL_CAL_ASIL_B_END}
PREDECL_CAL_C = {'START': PREDECL_CAL_ASIL_C_START, 'END': PREDECL_CAL_ASIL_C_END}
PREDECL_CAL_D = {'START': PREDECL_CAL_ASIL_D_START, 'END': PREDECL_CAL_ASIL_D_END}
PREDECL_CAL_EXT = {'START': PREDECL_CAL_EXT_START, 'END': PREDECL_CAL_EXT_END}
PREDECL_CAL_MERG = {'START': PREDECL_CAL_MERG_START, 'END': PREDECL_CAL_MERG_END}

PREDECL_CONST_QM = {'START': PREDECL_CONST_START, 'END': PREDECL_CONST_END}
PREDECL_CONST_A = {'START': PREDECL_CONST_ASIL_A_START, 'END': PREDECL_CONST_ASIL_A_END}
PREDECL_CONST_B = {'START': PREDECL_CONST_ASIL_B_START, 'END': PREDECL_CONST_ASIL_B_END}
PREDECL_CONST_C = {'START': PREDECL_CONST_ASIL_C_START, 'END': PREDECL_CONST_ASIL_C_END}
PREDECL_CONST_D = {'START': PREDECL_CONST_ASIL_D_START, 'END': PREDECL_CONST_ASIL_D_END}
PREDECL_CONST_EXT = {'START': PREDECL_CONST_EXT_START, 'END': PREDECL_CONST_EXT_END}
PREDECL_CONST_MERG = {'START': PREDECL_CONST_MERG_START, 'END': PREDECL_CONST_MERG_END}

PREDECL_NVM = {'START': PREDECL_NVM_START, 'END': PREDECL_NVM_END}
PREDECL_NVM_P = {'START': PREDECL_NVM_P_START, 'END': PREDECL_NVM_P_END}

CVC_SECURE = {
    'CODE': CVC_CODE_QM,
    'DISP': CVC_DISP_SECURE,
    'CAL': CVC_CAL_QM,
    'CONST': CVC_CONST_QM
}

CVC_ASIL_QM = {
    'CODE': CVC_CODE_QM,
    'DISP': CVC_DISP_QM,
    'CAL': CVC_CAL_QM,
    'CONST': CVC_CONST_QM
}

CVC_ASIL_A = {
    'CODE': CVC_CODE_A,
    'DISP': CVC_DISP_A,
    'CAL': CVC_CAL_A,
    'CONST': CVC_CONST_A
}

CVC_ASIL_B = {
    'CODE': CVC_CODE_B,
    'DISP': CVC_DISP_B,
    'CAL': CVC_CAL_B,
    'CONST': CVC_CONST_B
}

CVC_ASIL_C = {
    'CODE': CVC_CODE_C,
    'DISP': CVC_DISP_C,
    'CAL': CVC_CAL_C,
    'CONST': CVC_CONST_C
}

CVC_ASIL_D = {
    'CODE': CVC_CODE_D,
    'DISP': CVC_DISP_D,
    'CAL': CVC_CAL_D,
    'CONST': CVC_CONST_D
}

PREDECL_SECURE = {
    'CODE': PREDECL_CODE_QM,
    'DISP': PREDECL_DISP_SECURE,
    'CAL': PREDECL_CAL_QM,
    'CONST': PREDECL_CONST_QM
}

PREDECL_ASIL_QM = {
    'CODE': PREDECL_CODE_QM,
    'DISP': PREDECL_DISP_QM,
    'CAL': PREDECL_CAL_QM,
    'CONST': PREDECL_CONST_QM
}

PREDECL_ASIL_A = {
    'CODE': PREDECL_CODE_A,
    'DISP': PREDECL_DISP_A,
    'CAL': PREDECL_CAL_A,
    'CONST': PREDECL_CONST_A
}

PREDECL_ASIL_B = {
    'CODE': PREDECL_CODE_B,
    'DISP': PREDECL_DISP_B,
    'CAL': PREDECL_CAL_B,
    'CONST': PREDECL_CONST_B
}

PREDECL_ASIL_C = {
    'CODE': PREDECL_CODE_C,
    'DISP': PREDECL_DISP_C,
    'CAL': PREDECL_CAL_C,
    'CONST': PREDECL_CONST_C
}

PREDECL_ASIL_D = {
    'CODE': PREDECL_CODE_D,
    'DISP': PREDECL_DISP_D,
    'CAL': PREDECL_CAL_D,
    'CONST': PREDECL_CONST_D
}

PREDECL_EXTRA = {
    'NORM': PREDECL,
    'CAL_EXT': PREDECL_CAL_EXT,
    'CONST_EXT': PREDECL_CONST_EXT,
    'CAL_MERG': PREDECL_CAL_MERG,
    'CONST_MERG': PREDECL_CONST_MERG
}

NVM = {'CVC': CVC_NVM, 'PREDECL': PREDECL_NVM}
NVM_P = {'CVC': CVC_NVM_P, 'PREDECL': PREDECL_NVM_P}
SECURE = {'CVC': CVC_SECURE, 'PREDECL': PREDECL_SECURE}
ASIL_QM = {'CVC': CVC_ASIL_QM, 'PREDECL': PREDECL_ASIL_QM}
ASIL_A = {'CVC': CVC_ASIL_A, 'PREDECL': PREDECL_ASIL_A}
ASIL_B = {'CVC': CVC_ASIL_B, 'PREDECL': PREDECL_ASIL_B}
ASIL_C = {'CVC': CVC_ASIL_C, 'PREDECL': PREDECL_ASIL_C}
ASIL_D = {'CVC': CVC_ASIL_D, 'PREDECL': PREDECL_ASIL_D}

NVM_LEVEL_MAP = {
    'NORMAL': NVM,
    'PROTECTED': NVM_P
}

CVC_ASIL_LEVEL_MAP = {
    'QM': CVC_ASIL_QM,
    'A': CVC_ASIL_A,
    'B': CVC_ASIL_B,
    'C': CVC_ASIL_C,
    'D': CVC_ASIL_D
}

PREDECL_ASIL_LEVEL_MAP = {
    'QM': PREDECL_ASIL_QM,
    'A': PREDECL_ASIL_A,
    'B': PREDECL_ASIL_B,
    'C': PREDECL_ASIL_C,
    'D': PREDECL_ASIL_D
}

ASIL_LEVEL_MAP = {
    'QM': ASIL_QM,
    'A': ASIL_A,
    'B': ASIL_B,
    'C': ASIL_C,
    'D': ASIL_D
}



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Script to replace pragmas in hand written c-code."""

import re
import os
import sys
import shutil

REPO_ROOT = os.path.join(os.path.dirname(__file__), '..', '..')


class CodeReplacer:
    """Class to replace code in hand written c-code."""

    def replace_line(self, line):
        """Replace line."""
        raise NotImplementedError

    def replace_file(self, file_name):
        """Go through all lines in the file and replace pragmas."""
        tmp_file = file_name + '.tmp'
        with open(file_name, 'r', encoding='ascii', errors='ignore') as old_file:
            with open(tmp_file, 'w', encoding='ascii', errors='ignore') as new_file:
                for line in old_file.readlines():
                    line = self.replace_line(line)
                    new_file.write(line)
        shutil.move(tmp_file, file_name)


class PragmaReplacer(CodeReplacer):
    """Class to replace pragmas in hand written c-code."""

    def __init__(self):
        """Init."""
        self.cvc_started = False
        self.cvc = None
        self.regex = re.compile(r'^\s*#pragma section\s*(CVC(?P<cvc>[a-zA-Z0-9_]*))*\s*$')
        self.template = '#include "CVC_{cvc}_{start_or_end}.h"\n'

    def replace_line(self, line):
        """Replace line (if it has a pragma)."""
        match = self.regex.match(line)
        if match:
            if self.cvc_started:
                line = self.template.format(cvc=self.cvc,
                                            start_or_end='END')
                self.cvc = None
            else:
                self.cvc = match.group('cvc') or 'CODE'
                line = self.template.format(cvc=self.cvc,
                                            start_or_end='START')
            self.cvc_started = not self.cvc_started
        return line


class CodeSwitchReplacer(CodeReplacer):
    """Class to replace code switch includes in hand written c-code."""

    def __init__(self):
        """Init."""
        self.regex = re.compile(r'(.*)SPM_Codeswitch_Setup(_PVC)?(.*)')
        self.template = '{}VcCodeSwDefines{}\n'

    def replace_line(self, line):
        """Replace include code switch file."""
        match = self.regex.match(line)
        if match:
            return self.template.format(match.group(1), match.group(3))
        return line


def update_files(files):
    """Replace code in handwritten file."""
    for file_path in files:
        PragmaReplacer().replace_file(file_path)
        CodeSwitchReplacer().replace_file(file_path)


def update_test_files(files):
    """Replace code in handwritten file."""
    for file_path in files:
        CodeSwitchReplacer().replace_file(file_path)


def get_files_to_update(source_dir_name):
    """Get files to update."""
    files = [os.path.join('./Models/SSPCECD/VcTqReq__DIESEL', source_dir_name, 'invTab2_UInt16_func.c'),
             os.path.join('./Models/SSPCECD/VcCmbNOx__DIESEL', source_dir_name, 'NNEval15x8_func.c'),
             os.path.join('./Models/SSPCECD/VcTqEff__DIESEL', source_dir_name, 'NNEval15x8_func.c')]
    for path, _, filenames in os.walk(os.path.join(REPO_ROOT, 'Models', 'SSPDL')):
        if path.endswith(os.path.join('Mdl', source_dir_name)) or\
                path.endswith(os.path.join('Mdl__denso', source_dir_name)):
            for filename in filenames:
                if (filename.endswith('.c') and not filename.endswith('Mdl.c')) or \
                   (filename.endswith('.h') and not filename.endswith('Mdl.h')):
                    file_path = os.path.relpath(os.path.join(path, filename))
                    files.append(file_path)
    return files


def get_test_files_to_update():
    """Get files to update."""
    files = []
    for path, _, filenames in os.walk(os.path.join(REPO_ROOT, 'Models')):
        for filename in filenames:
            if filename.endswith('.py'):
                file_path = os.path.relpath(os.path.join(path, filename))
                files.append(file_path)
    return files


def main():
    """Replace incorrect pragmas in handwritten c-code."""
    files_to_update = get_files_to_update('pybuild_src')
    update_files(files_to_update)
    test_files_to_update = get_test_files_to_update()
    update_test_files(test_files_to_update)
    return 0


if __name__ == '__main__':
    sys.exit(main())



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Module containing classes for generation of scheduling call points."""
from os.path import join as pjoin
from os.path import split as psplit

import powertrain_build.build_defs as build_defs
from powertrain_build.build_proj_config import BuildProjConfig
from powertrain_build.problem_logger import ProblemLogger
from powertrain_build.unit_configs import UnitConfigs


class SchedFuncs(ProblemLogger):
    """Generate scheduling functions for all unit initialisation and unit scheduling functions.

    Generates VcExtINI.c containing one function, VcExtINI, that calls all included units' initialisation
    (RESTART) functions.

    Generates one VcExt*.c file per configured scheduling raster containing one function collecting
    all unit scheduling functions for the specified time-raster.

    TODO: Add the debug functions
    """

    # Not included in VcExtINI.c and VcUnitTsDefines.h.
    RESTART_FNC_EXCLUDE = ('VcDebug', 'VcDebugOutput', 'VcDebugSafe', 'VcDebugOutputSafe')

    def __init__(self, build_proj_cfg, unit_cfg, dbg_dict):
        """Constructor.

        Args:
            build_proj_cfg (BuildProjConfig): project configuration.
            unit_cfg (UnitConfigs): Class holding all unit interfaces.
            dbg_dict (dict): Dictionary containing debug information, keys in RESTART_FNC_EXCLUDE.
        """
        super().__init__()

        if isinstance(build_proj_cfg, BuildProjConfig) and isinstance(unit_cfg, UnitConfigs):
            self._prj_cfg = build_proj_cfg
            self._unit_cfg = unit_cfg
            self._dbg_dict = dbg_dict
        else:
            err = (
                'Input arguments should be an instance of:'
                f'BuildProjConfig, not {type(build_proj_cfg)}'
                f'AND/OR UnitConfigs, not {type(unit_cfg)}'
            )
            raise TypeError(err)

    def generate_sched_c_fncs(self, generate_rte_checkpoint_calls):
        """Generate a c-files.

        One per scheduling raster and
        one for the running the Init functions

        TODO: Do the suppliers need a header per scheduling function?

        Args:
            generate_rte_checkpoint_calls (bool): Generate RTE function checkpoint calls.
        """
        dst_dir = self._prj_cfg.get_src_code_dst_dir()
        unit_raster_cfg = self._prj_cfg.get_units_raster_cfg()
        per_unit_cfg = self._unit_cfg.get_per_unit_cfg()
        prefix = self._prj_cfg.get_scheduler_prefix()
        unit_counter = 1
        rte_checkpoint_id_size = self._prj_cfg.get_code_generation_config("rteCheckpointIdSize")
        if rte_checkpoint_id_size is None:
            if generate_rte_checkpoint_calls:
                self.warning("RTE checkpoint ID size not set, assuming size is the same as the number of units.")
            rte_checkpoint_id_size = 0
            for units_in_raster in unit_raster_cfg['Rasters'].values():
                rte_checkpoint_id_size += len(units_in_raster)
        include_name = self._prj_cfg.get_composition_config("softwareComponentName")
        function_name = self._prj_cfg.get_composition_config("compositionName")
        if function_name is None:
            function_name = include_name

        init_ext_def_s = ""
        init_call_s = f'#include "{build_defs.CVC_CODE_START}"\n\n'
        init_call_s += f'void {prefix}VcExtINI(void)\n{{\n'
        for raster in unit_raster_cfg['SampleTimes']:
            if raster == "init":
                for unit in unit_raster_cfg['Rasters'][raster]:
                    if unit not in SchedFuncs.RESTART_FNC_EXCLUDE and 'code_generator' in per_unit_cfg[unit]:
                        code_generator = per_unit_cfg[unit]['code_generator']
                    else:
                        code_generator = 'target_link'

                    unit_name = unit.rsplit('__', 1)[0]
                    if code_generator == 'embedded_coder':
                        init_ext_def_s += f'extern void {unit_name}_initialize(void);\n'
                    else:
                        init_ext_def_s += f'extern void RESTART_{unit_name}(void);\n'

                    if unit_name not in SchedFuncs.RESTART_FNC_EXCLUDE:
                        if code_generator == 'embedded_coder':
                            init_call_s += f'  {unit_name}_initialize();\n'
                        else:
                            init_call_s += f'  RESTART_{unit_name}();\n'
            else:
                includes = ""
                rteCheckpointEntityName = self._prj_cfg.get_code_generation_config(
                    "customRteCheckpointEntityName"
                ).format(raster=raster)
                if generate_rte_checkpoint_calls:
                    includes += f'#include "Rte_{include_name}.h"\n'
                if includes:
                    includes += '\n'
                ext_def_str = ""
                call_str = f'#include "{build_defs.CVC_CODE_START}"\n'
                call_str += f'void {prefix}{raster}(void)\n{{\n'
                for unit in unit_raster_cfg['Rasters'][raster]:
                    if unit_counter > rte_checkpoint_id_size:
                        self.critical(
                            'RTE checkpoint ID size %d is smaller than the number of units.', rte_checkpoint_id_size
                        )
                    if unit not in SchedFuncs.RESTART_FNC_EXCLUDE and 'code_generator' in per_unit_cfg[unit]:
                        code_generator = per_unit_cfg[unit]['code_generator']
                    else:
                        code_generator = 'target_link'

                    unit_name = unit.rsplit('__', 1)[0]
                    if code_generator == 'embedded_coder':
                        if unit_name not in SchedFuncs.RESTART_FNC_EXCLUDE:
                            init_ext_def_s += f'extern void {unit_name}_initialize(void);\n'
                            init_call_s += f'  {unit_name}_initialize();\n'
                            ext_def_str += f'extern void {unit_name}_step(void);\n'
                            call_str += f'  {unit_name}_step();\n'
                        elif self._dbg_dict[unit_name]:
                            ext_def_str += f'extern void {unit_name}_step(void);\n'
                            call_str += f'  {unit_name}_step();\n'
                    else:
                        if generate_rte_checkpoint_calls:
                            call_str += (
                                '  (void)Rte_Call_alive_WdgMSupervisedEntity_'
                                f'{rteCheckpointEntityName}({unit_counter + 1}U);\n'  # Function IDs start at 2
                            )
                        if unit_name not in SchedFuncs.RESTART_FNC_EXCLUDE:
                            init_ext_def_s += f'extern void RESTART_{unit_name}(void);\n'
                            init_call_s += f'  RESTART_{unit_name}();\n'
                            ext_def_str += f'extern void {unit_name}(void);\n'
                            call_str += f'  {unit_name}();\n'
                        elif self._dbg_dict[unit_name]:
                            ext_def_str += f'extern void {unit_name}(void);\n'
                            call_str += f'  {unit_name}();\n'
                    unit_counter += 1
                if generate_rte_checkpoint_calls:
                    # Also call remaining dummy IDs
                    for i in range(unit_counter, rte_checkpoint_id_size + 1):
                        call_str += (
                            '  (void)Rte_Call_alive_WdgMSupervisedEntity_'
                            f'{rteCheckpointEntityName}({i + 1}U);\n'  # Function IDs start at 2 (same as above)
                        )
                ext_def_str += '\n'
                call_str += '}\n'
                call_str += f'#include "{build_defs.CVC_CODE_END}"\n\n'

                file_name = pjoin(dst_dir, f'{raster}.c')
                with open(file_name, 'w', encoding="utf-8") as fhndl:
                    fhndl.write(includes)
                    fhndl.write(ext_def_str)
                    fhndl.write(call_str)
                    self.info(f'generated {file_name}')
                file_name = pjoin(dst_dir, f'{raster}.h')
                with open(file_name, 'w', encoding="utf-8") as f_h:
                    f_h.write(f'#ifndef {raster.upper()}_H\n')
                    f_h.write(f'#define {raster.upper()}_H\n\n')
                    f_h.write(f'void {prefix}{raster}(void);\n\n')
                    f_h.write(f'#endif //{raster.upper()}_H\n')
        init_ext_def_s += '\n'
        init_call_s += '}\n\n'
        init_call_s += f'#include "{build_defs.CVC_CODE_END}"\n\n'
        file_name = pjoin(dst_dir, 'VcExtINI.c')
        with open(file_name, 'w', encoding="utf-8") as fhndl:
            fhndl.write(init_ext_def_s)
            fhndl.write(init_call_s)
            self.info('generated %s', file_name)

        file_name = pjoin(dst_dir, 'VcExtINI.h')
        with open(file_name, 'w', encoding="utf-8") as f_h:
            f_h.write('#ifndef VCEXTINI_H\n')
            f_h.write('#define VCEXTINI_H\n')
            f_h.write(f'void {prefix}VcExtINI(void);\n')
            f_h.write('#endif\n')

    def generate_ts_defines(self, file_name):
        """Generate the ts defines.

        Generates defines needed for all units, and
        writes it to a file with the supplied filename.

        Args:
            file_name (str): ts header file name
        """
        with open(file_name, 'w', encoding="utf-8") as fhndl:
            fhndl.write('/* Autogenerated by build system */\n')
            def_name = psplit(file_name)[1].replace('.', '_').upper()
            fhndl.write(f'#ifndef {def_name}\n')
            fhndl.write(f'#define {def_name}\n')
            unit_raster_cfg = self._prj_cfg.get_units_raster_cfg()
            for raster in unit_raster_cfg['SampleTimes']:
                fhndl.write(f'\n/**** raster {raster} *****/\n\n')
                for unit in unit_raster_cfg['Rasters'][raster]:
                    if unit not in SchedFuncs.RESTART_FNC_EXCLUDE:
                        t_sample = unit_raster_cfg['SampleTimes'][raster]
                        fhndl.write(f'#define ts_{unit} ((Float32) {t_sample}F)\n')
            fhndl.write(f'#endif /* {def_name} */\n')
            self.info('generated %s', file_name)
